// Copyright 2015 Open Source Robotics Foundation, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef PRIORITY_EXECUTOR__PRIORITY_MEMORY_STRATEGY_HPP_
#define PRIORITY_EXECUTOR__PRIORITY_MEMORY_STRATEGY_HPP_

#include <cstdint>
#include <deque>
#include <memory>
#include <semaphore.h>
#include <set>
#include <time.h>
#include <vector>

#include "rcl/allocator.h"

#include "rclcpp/allocator/allocator_common.hpp"
#include "rclcpp/memory_strategy.hpp"

#include "simple_timer/rt-sched.hpp"

/// Delegate for handling memory allocations while the Executor is executing.
/**
 * By default, the memory strategy dynamically allocates memory for structures
 * that come in from the rmw implementation after the executor waits for work,
 * based on the number of entities that come through.
 */

namespace priority_executor {

enum class ExecutableType {
    SUBSCRIPTION,
    SERVICE,
    CLIENT,
    TIMER,
    WAITABLE
};

std::ostream& operator<<(std::ostream& os, const ExecutableType& obj);

enum class ExecutableScheduleType {
    DEADLINE = 0,
    CHAIN_AWARE_PRIORITY,
    CHAIN_INDEPENDENT_PRIORITY,  // not used here
    DEFAULT,                     // not used here
};

std::ostream& operator<<(std::ostream& os, const ExecutableScheduleType& obj);

class PriorityExecutable {
public:
    static size_t num_executables;

    explicit PriorityExecutable(
        std::shared_ptr<const void> h,
        int p,
        ExecutableType t,
        ExecutableScheduleType sched_type = ExecutableScheduleType::CHAIN_INDEPENDENT_PRIORITY);

    PriorityExecutable();

    void dont_run();
    void allow_run();
    void increment_counter();
    bool operator==(PriorityExecutable const &other) const;
    friend std::ostream& operator<<(std::ostream& os, PriorityExecutable const &pe);

    std::shared_ptr<const void> handle;
    ExecutableType type;
    bool can_be_run = false;
    std::shared_ptr<rclcpp::Waitable> waitable;
    ExecutableScheduleType sched_type;

    int priority = 0;
    long period = 1000;  // milliseconds

    bool is_first_in_chain = false;
    bool is_last_in_chain = false;
    std::deque<uint64_t>* deadlines = nullptr;  // chain aware deadlines
    std::shared_ptr<rclcpp::TimerBase> timer_handle;
    int chain_id = 0;  // just used for logging
    int counter = 0;   // chain aware priority
    int executable_id = 0;
    std::string name = "";
};

class PriorityExecutableComparator {
public:
    bool operator()(PriorityExecutable const *p1, PriorityExecutable const *p2) const;
};

template <typename Alloc = std::allocator<void>>
class PriorityMemoryStrategy : public rclcpp::memory_strategy::MemoryStrategy {
public:
    RCLCPP_SMART_PTR_DEFINITIONS(PriorityMemoryStrategy<Alloc>)

    using VoidAllocTraits = typename rclcpp::allocator::AllocRebind<void*, Alloc>;
    using VoidAlloc = typename VoidAllocTraits::allocator_type;

    explicit PriorityMemoryStrategy(std::shared_ptr<Alloc> allocator);
    PriorityMemoryStrategy();

    node_time_logger logger_;

    // Override methods
    void add_guard_condition(const rcl_guard_condition_t* guard_condition) override;
    void remove_guard_condition(const rcl_guard_condition_t* guard_condition) override;
    void clear_handles() override;
    void remove_null_handles(rcl_wait_set_t* wait_set) override;
    bool collect_entities(const WeakNodeList& weak_nodes) override;
    void add_waitable_handle(const rclcpp::Waitable::SharedPtr& waitable) override;
    bool add_handles_to_wait_set(rcl_wait_set_t* wait_set) override;

    // Class-specific methods
    void get_next_executable(rclcpp::AnyExecutable& any_exec, const WeakNodeList& weak_nodes);
    void post_execute(rclcpp::AnyExecutable any_exec, int thread_id = -1);

    // Override virtual methods
    void get_next_subscription(rclcpp::AnyExecutable& any_exec, const WeakNodeList& weak_nodes) override;
    void get_next_service(rclcpp::AnyExecutable& any_exec, const WeakNodeList& weak_nodes) override;
    void get_next_client(rclcpp::AnyExecutable& any_exec, const WeakNodeList& weak_nodes) override;
    void get_next_timer(rclcpp::AnyExecutable& any_exec, const WeakNodeList& weak_nodes) override;
    void get_next_waitable(rclcpp::AnyExecutable& any_exec, const WeakNodeList& weak_nodes) override;

    // Priority handling methods
    std::shared_ptr<PriorityExecutable> get_priority_settings(std::shared_ptr<const void> executable);
    rcl_allocator_t get_allocator() override;

    // Counter methods
    size_t number_of_ready_subscriptions() const override;
    size_t number_of_ready_services() const override;
    size_t number_of_ready_events() const override;
    size_t number_of_ready_clients() const override;
    size_t number_of_guard_conditions() const override;
    size_t number_of_ready_timers() const override;
    size_t number_of_waitables() const override;

    // Executable priority methods
    void set_executable_priority(std::shared_ptr<const void> handle, int priority, ExecutableType t);
    void set_executable_priority(std::shared_ptr<const void> handle, int priority, ExecutableType t, 
                               ExecutableScheduleType sc, int chain_index);
    void set_executable_deadline(std::shared_ptr<const void> handle, int period, ExecutableType t, 
                               int chain_id = 0, std::string name = "");
    int get_priority(std::shared_ptr<const void> executable);
    void set_first_in_chain(std::shared_ptr<const void> exec_handle);
    void set_last_in_chain(std::shared_ptr<const void> exec_handle);
    void assign_deadlines_queue(std::shared_ptr<const void> exec_handle, std::deque<uint64_t>* deadlines);
    std::shared_ptr<std::deque<uint64_t>> get_chain_deadlines(int chain_id);

private:
    std::shared_ptr<PriorityExecutable> get_and_reset_priority(std::shared_ptr<const void> executable, 
                                                              ExecutableType t);

    template <typename T>
    using VectorRebind = std::vector<T, typename std::allocator_traits<Alloc>::template rebind_alloc<T>>;

    // Member variables
    VectorRebind<const rcl_guard_condition_t*> guard_conditions_;
    VectorRebind<std::shared_ptr<const rcl_subscription_t>> subscription_handles_;
    VectorRebind<std::shared_ptr<const rcl_service_t>> service_handles_;
    VectorRebind<std::shared_ptr<const rcl_client_t>> client_handles_;
    VectorRebind<std::shared_ptr<const rcl_timer_t>> timer_handles_;
    VectorRebind<std::shared_ptr<rclcpp::Waitable>> waitable_handles_;
    std::shared_ptr<VoidAlloc> allocator_;
    std::map<std::shared_ptr<const void>, std::shared_ptr<PriorityExecutable>> priority_map;
    std::map<int, std::shared_ptr<std::deque<uint64_t>>> chain_deadlines;
    std::set<const PriorityExecutable*, PriorityExecutableComparator> all_executables_;

    void add_executable_to_queue(std::shared_ptr<PriorityExecutable> p);
};

}  // namespace priority_executor

#endif  // PRIORITY_EXECUTOR__PRIORITY_MEMORY_STRATEGY_HPP_